# coding=utf-8
import os
# Use GPU0
from tf_geometric.utils import tf_utils

os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
import tensorflow as tf
import tf_geometric as tfg
import tf_sparse as tfs


num_nodes = 5
edge_index = [
    [0, 0, 1, 2, 3, 3, 4],
    [1, 2, 4, 3, 1, 4, 2]
]

# Sparse node features
# tf.sparse.eye creates a two-dimensional sparse tensor with ones along the diagonal
# x is the one-hot encoding of node ids (from 0 to num_nodes - 1) in the form of a sparse matrix
# This is usually used for feature-less cases, such as recommendation systems.
x = tfs.eye(num_nodes)
print("Sparse (One-hot) Node Features: ")
print(x.to_dense())

# tf.sparse.SparseTensor can be used as node features (x)
graph = tfg.Graph(x, edge_index).convert_edge_to_directed()
print("\nConstructed Graph:")
print(graph)

# create a one-layer GNN model
model = tfg.layers.GCN(4)
# model = tfg.layers.SGC(4, k=3)
# model = tfg.layers.ChebyNet(4, k=4)
# model = tfg.layers.TAGCN(4, k=4)
# model = tfg.layers.APPNP([4, 4], tf.nn.relu, num_iterations=10)

# predict with the GCN model
@tf_utils.function
def forward(graph):
    return model([graph.x, graph.edge_index])

logits = forward(graph)
print("\nModel Output:")
print(logits)

# tfg.Graph objects with sparse node features can also be combined into a tfg.BatchGraph object
batch_graph = tfg.BatchGraph.from_graphs([graph, graph])
print("\nCombined Batch Graph")
print(batch_graph)


x = tf.sparse.eye(10)
# x = tfs.eye(10)
g = tfg.Graph(x, [[0, 5, 2], [5, 6, 4]])

# print(tfg.BatchGraph.from_graphs([g, g, g]).x.to_dense())

print(tf.sparse.to_dense(tfg.BatchGraph.from_graphs([g, g, g]).x))

# new_g = g.sample_new_graph_by_node_index([0, 5, 6])
#
# print(new_g.x.to_dense())
#
# # print(tf.sparse.to_dense(new_g.x))
# print(new_g.edge_index)



# coding=utf-8
import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
from tf_geometric.utils import tf_utils
import tf_geometric as tfg
import tensorflow as tf
import time
from tqdm import tqdm
import tf_sparse as tsp

graph, (train_index, valid_index, test_index) = tfg.datasets.CoraDataset().load_data()
graph.x = tsp.eye(graph.num_nodes)
num_classes = graph.y.max() + 1
drop_rate = 0.6


# Multi-layer GAT Model
class GATModel(tf.keras.Model):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gat0 = tfg.layers.GAT(64, activation=tf.nn.relu, num_heads=8, attention_units=8, drop_rate=drop_rate)
        self.gat1 = tfg.layers.GAT(num_classes, num_heads=1, attention_units=1, drop_rate=drop_rate)

        # The GAT paper mentioned that: "Specially, if we perform multi-head attention on the final (prediction) layer of
        # the network, concatenation is no longer sensible - instead, we employ averaging".
        # In tf_geometric, if you want to set num_heads > 1 for the last output GAT layer, you can set split_value_heads=False
        # as follows to employ averaging instead of concatenation.
        # self.gat1 = tfg.layers.GAT(num_classes, num_heads=8, attention_units=8, split_value_heads=False, drop_rate=drop_rate)

        # self.dropout = tf.keras.layers.Dropout(drop_rate)
        self.dropout = tsp.layers.Dropout(drop_rate)

    def call(self, inputs, training=None, mask=None, cache=None):
        x, edge_index = inputs
        h = self.dropout(x, training=training)
        h = self.gat0([h, edge_index], training=training)
        h = self.dropout(h, training=training)
        h = self.gat1([h, edge_index], training=training)
        return h


model = GATModel()


# @tf_utils.function can speed up functions for TensorFlow 2.x
@tf_utils.function
def forward(graph, training=False):
    return model([graph.x, graph.edge_index], training=training)


@tf_utils.function
def compute_loss(logits, mask_index, vars):
    masked_logits = tf.gather(logits, mask_index)
    masked_labels = tf.gather(graph.y, mask_index)
    losses = tf.nn.softmax_cross_entropy_with_logits(
        logits=masked_logits,
        labels=tf.one_hot(masked_labels, depth=num_classes)
    )

    kernel_vars = [var for var in vars if "kernel" in var.name]
    l2_losses = [tf.nn.l2_loss(kernel_var) for kernel_var in kernel_vars]

    return tf.reduce_mean(losses) + tf.add_n(l2_losses) * 5e-4


@tf_utils.function
def train_step():
    with tf.GradientTape() as tape:
        logits = forward(graph, training=True)
        loss = compute_loss(logits, train_index, tape.watched_variables())

    vars = tape.watched_variables()
    grads = tape.gradient(loss, vars)
    optimizer.apply_gradients(zip(grads, vars))
    return loss


@tf_utils.function
def evaluate():
    logits = forward(graph)
    masked_logits = tf.gather(logits, test_index)
    masked_labels = tf.gather(graph.y, test_index)
    y_pred = tf.argmax(masked_logits, axis=-1, output_type=tf.int32)

    corrects = tf.equal(y_pred, masked_labels)
    accuracy = tf.reduce_mean(tf.cast(corrects, tf.float32))
    return accuracy


optimizer = tf.keras.optimizers.Adam(learning_rate=5e-3)

for step in range(1, 401):
    loss = train_step()

    if step % 20 == 0:
        accuracy = evaluate()
        print("step = {}\tloss = {}\taccuracy = {}".format(step, loss, accuracy))

print("\nstart speed test...")
num_test_iterations = 1000
start_time = time.time()
for _ in tqdm(range(num_test_iterations)):
    logits = forward(graph)
end_time = time.time()
print("mean forward time: {} seconds".format((end_time - start_time) / num_test_iterations))

if tf.__version__[0] == "1":
    print("** @tf_utils.function is disabled in TensorFlow 1.x. "
          "Upgrade to TensorFlow 2.x for 10X faster speed. **")



# coding=utf-8
import tensorflow as tf
import numpy as np


def drop_edge(inputs, rate=0.5, force_undirected=False, training=None):
    """

    :param inputs: List of edge_index and other edge attributes [edge_index, edge_attr, ...]
    :param rate: dropout rate
    :param force_undirected: If set to `True`, will either
            drop or keep both edges of an undirected edge.
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: List of dropped edge_index and other dropped edge attributes
    """

    if not training:
        return inputs

    if rate < 0.0 or rate > 1.0:
        raise ValueError('Dropout probability has to be between 0 and 1, '
                         'but got {}'.format(rate))

    edge_index, *edge_attrs = inputs

    edge_index_is_tensor = tf.is_tensor(edge_index)
    if not edge_index_is_tensor:
        edge_index = tf.convert_to_tensor(edge_index)

    row, col = edge_index[0], edge_index[1]
    if force_undirected:
        index = tf.where(tf.less(row, col))
        index = tf.boolean_mask(index, tf.greater(tf.nn.dropout(tf.ones_like(index, dtype=tf.float32), rate), 0))
        dropped_edge_index = tf.gather(edge_index, index, axis=-1)
        dropped_edge_index = tf.concat([dropped_edge_index, tf.gather(dropped_edge_index, [1, 0])], axis=-1)
        index = tf.concat([index, index], axis=-1)
    else:
        index = tf.boolean_mask(tf.range(0, tf.shape(row)[0]),
                                tf.greater(tf.nn.dropout(tf.ones_like(row, dtype=tf.float32), rate), 0))
        dropped_edge_index = tf.gather(edge_index, index, axis=-1)

    for i in range(len(edge_attrs)):
        if tf.is_tensor(edge_attrs[i]):
            edge_attrs[i] = tf.gather(edge_attrs[i], index, axis=-1)
        else:
            edge_attrs[i] = np.take(edge_attrs[i], index, axis=-1)

    if not edge_index_is_tensor:
        dropped_edge_index = dropped_edge_index.numpy()

    return [dropped_edge_index] + edge_attrs



# coding=utf-8
import os
import tensorflow as tf
from tf_geometric.utils import tf_utils

import tf_geometric as tfg
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

class A():

    def __init__(self, edge_index):
        self.edge_index = edge_index

    @property
    def num_edges(self):
        shape = tf.shape(self.edge_index)
        if shape[0] == 0:
            return 10
        else:
            return 6

@tf_utils.function
def test(x, edge_index):
    # a = A(edge_index)
    # return a.num_edges()
    # return tf.shape(edge_index)[0] == 0
    graph = tfg.Graph(x, edge_index)
    return graph.num_edges


x = tf.random.truncated_normal([10, 5])
edge_index = tf.convert_to_tensor([
    [0, 1, 3, 5],
    [0, 2, 4, 5]
])

print(test(x, edge_index))


# coding=utf-8
from tf_geometric.datasets.csr_npz import CSRNPZDataset


class CoauthorDataset(CSRNPZDataset):

    def __init__(self, dataset_name, dataset_root_path=None):
        """

        :param dataset_name: "coauthor-cs" | "coauthor-physics"
        :param dataset_root_path:
        """
        super().__init__(dataset_name=dataset_name,
                         download_urls=[
                             "https://github.com/CrawlScript/gnn_datasets/raw/master/Coauthor/{}.zip".format(dataset_name),
                             "http://cdn.zhuanzhi.ai/github/{}.zip".format(dataset_name)
                         ],
                         download_file_name="{}.zip".format(dataset_name),
                         cache_name=None,
                         dataset_root_path=dataset_root_path,
                         )


class CoauthorCSDataset(CoauthorDataset):

    def __init__(self, dataset_root_path=None):
        super().__init__("coauthor-cs", dataset_root_path=dataset_root_path)


class CoauthorPhysicsDataset(CoauthorDataset):

    def __init__(self, dataset_root_path=None):
        super().__init__("coauthor-physics", dataset_root_path=dataset_root_path)



# coding=utf-8
import tensorflow as tf

from tf_geometric.sparse.sparse_adj import SparseAdj
from tf_geometric.utils.graph_utils import add_self_loop_edge
import tf_sparse as tfs


# follow Transformer-Style Attention
# Attention is all you need
def gat(x, edge_index,
        query_kernel, query_bias, query_activation,
        key_kernel, key_bias, key_activation,
        kernel, bias=None, activation=None, num_heads=1,
        split_value_heads=True, drop_rate=0.0, training=False):
    """

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param query_kernel: Tensor, shape: [num_features, num_query_features], weight for Q in attention
    :param query_bias: Tensor, shape: [num_query_features], bias for Q in attention
    :param query_activation: Activation function for Q in attention.
    :param key_kernel: Tensor, shape: [num_features, num_key_features], weight for K in attention
    :param key_bias: Tensor, shape: [num_key_features], bias for K in attention
    :param key_activation: Activation function for K in attention.
    :param kernel: Tensor, shape: [num_features, num_output_features], weight
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param num_heads: Number of attention heads.
    :param split_value_heads: Boolean. If true, split V as value attention heads, and then concatenate them as output.
        Else, num_heads replicas of V are used as value attention heads, and the mean of them are used as output.
    :param drop_rate: Dropout rate.
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tfs.shape(x)[0]

    # self-attention
    edge_index, edge_weight = add_self_loop_edge(edge_index, num_nodes)

    row, col = edge_index[0], edge_index[1]

    x_is_sparse = isinstance(x, tf.sparse.SparseTensor)

    if x_is_sparse:
        Q = tf.sparse.sparse_dense_matmul(x, query_kernel)
    else:
        Q = x @ query_kernel
    Q += query_bias
    if query_activation is not None:
        Q = query_activation(Q)
    Q = tf.gather(Q, row)

    if x_is_sparse:
        K = tf.sparse.sparse_dense_matmul(x, key_kernel)
    else:
        K = x @ key_kernel
    K += key_bias
    if key_activation is not None:
        K = key_activation(K)
    K = tf.gather(K, col)

    if x_is_sparse:
        V = tf.sparse.sparse_dense_matmul(x, kernel)
    else:
        V = x @ kernel

    # xxxxx_ denotes the multi-head style stuff
    Q_ = tf.concat(tf.split(Q, num_heads, axis=-1), axis=0)
    K_ = tf.concat(tf.split(K, num_heads, axis=-1), axis=0)
    # splited queries and keys are modeled as virtual vertices
    qk_edge_index_ = tf.concat([edge_index + i * num_nodes for i in range(num_heads)], axis=1)

    scale = tf.math.sqrt(tf.cast(tf.shape(Q_)[-1], tf.float32))
    att_score_ = tf.reduce_sum(Q_ * K_, axis=-1) / scale

    # new implementation based on SparseAdj
    num_nodes_ = num_nodes * num_heads
    sparse_att_adj = SparseAdj(qk_edge_index_, att_score_, [num_nodes_, num_nodes_]) \
        .segment_softmax(axis=-1) \
        .dropout(drop_rate, training=training)

    if split_value_heads:
        V_ = tf.concat(tf.split(V, num_heads, axis=-1), axis=0)
    else:
        V_ = V
        edge_index_ = tf.tile(edge_index, [1, num_heads])
        sparse_att_adj = SparseAdj(edge_index_, sparse_att_adj.edge_weight, [num_nodes, num_nodes])

    h_ = sparse_att_adj @ V_

    # old implementation
    # normed_att_score_ = segment_softmax(att_score_, qk_edge_index_[0], num_nodes * num_heads)
    #
    # if training and drop_rate > 0.0:
    #     normed_att_score_ = tf.compat.v2.nn.dropout(normed_att_score_, drop_rate)
    #
    # if split_value_heads:
    #     V_ = tf.concat(tf.split(V, num_heads, axis=-1), axis=0)
    #     edge_index_ = qk_edge_index_
    # else:
    #     V_ = V
    #     edge_index_ = tf.tile(edge_index, [1, num_heads])
    #
    # h_ = aggregate_neighbors(
    #     V_, edge_index_, normed_att_score_,
    #     gcn_mapper,
    #     sum_reducer,
    #     identity_updater
    # )

    if split_value_heads:
        h = tf.concat(tf.split(h_, num_heads, axis=0), axis=-1)
    else:
        h = h_ / num_heads

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    return h



# coding=utf-8
from multiprocessing import Pool

from tf_geometric.data.graph import Graph
from tf_geometric.data.dataset import DownloadableDataset
import os
import numpy as np
import scipy.sparse as sp
from tqdm import tqdm


class ModelNetDataset(DownloadableDataset):
    def __init__(self, dataset_name, download_urls=None, dataset_root_path=None, num_processes=50):
        super().__init__(dataset_name, download_urls,
                         download_file_name="{}.zip".format(dataset_name),
                         cache_name="cache.p",
                         dataset_root_path=dataset_root_path)
        self.num_processes = num_processes

    def read_off(self, off_file_info):
        off_fpath, label_index = off_file_info
        with open(off_fpath, "r", encoding="utf-8") as f:
            # jump top line, usually "OFF\n"
            line = f.readline()

            # some files does not have "\n" after "OFF", and the number information is in the first line after "OFF"
            if line != "OFF\n":
                line = line[3:]
            else:
                # Usually, the number information is in the second line
                line = f.readline()

            # parse numbers
            num_nodes, num_faces, _ = [int(item) for item in line.split()]

            node_feature_matrix = []
            for _ in range(num_nodes):
                line = f.readline()
                node_features = [float(item) for item in line.split()]
                node_feature_matrix.append(node_features)

            triangles = []
            for _ in range(num_faces):
                line = f.readline()
                items = [int(item) for item in line.split()]
                num_face_nodes = items[0]

                # triangle
                if num_face_nodes == 3:
                    triangles.append(items[1:])
                # rectangle
                else:
                    triangles.append([items[1], items[2], items[3]])
                    triangles.append([items[1], items[3], items[4]])

        x = np.array(node_feature_matrix)
        triangles = np.array(triangles)
        edges = np.concatenate([
            triangles[:, :2],
            triangles[:, 1:],
            triangles[:, ::2]
        ], axis=0)
        row, col = edges[:, 0], edges[:, 1]
        row, col = np.concatenate([row, col], axis=0), np.concatenate([col, row], axis=0)

        adj = sp.csr_matrix((np.ones_like(row), (row, col)), shape=[num_nodes, num_nodes])
        adj.data[adj.data > 1] = 1

        adj = adj.tocoo()
        edge_index = np.stack([adj.row, adj.col], axis=0)

        graph = Graph(x=x, edge_index=edge_index, y=[label_index])

        return graph

    def process(self):
        data_dir = os.path.join(self.raw_root_path, self.dataset_name)
        sorted_sub_dir_names = sorted([sub_dir_name for sub_dir_name in os.listdir(data_dir)
                                       if os.path.isdir(os.path.join(data_dir, sub_dir_name))])
        label_names = sorted_sub_dir_names

        # label_name_index_dict = {label_name: label_index for label_index, label_name in enumerate(label_names)}

        train_graphs = []
        test_graphs = []

        for label_index, label_name in enumerate(label_names):
            print("processing {} ({}/{}):".format(label_name, label_index + 1, len(label_names)))

            label_path = os.path.join(data_dir, label_name)

            for split_index, split in enumerate(["train", "test"]):
                print("\treading {} graphs".format(split))

                split_graphs = train_graphs if split == "train" else test_graphs
                split_path = os.path.join(label_path, split)
                off_fpaths = [os.path.join(split_path, off_fname) for off_fname in os.listdir(split_path) if
                              off_fname != ".DS_Store"]

                pool_inputs = [[off_fpath, label_index] for off_fpath in off_fpaths]
                with Pool(processes=self.num_processes) as p:
                    with tqdm(total=len(off_fpaths)) as pbar:
                        for _, graph in enumerate(p.imap_unordered(self.read_off, pool_inputs)):
                            split_graphs.append(graph)
                            pbar.update()

                # split_path = os.path.join(label_path, split)
                # for off_fname in tqdm(os.listdir(split_path)):
                #     if off_fname == ".DS_Store":
                #         continue
                #     off_fpath = os.path.join(split_path, off_fname)
                #     graph = self.read_off(off_fpath, label_index)
                #     split_graphs.append(graph)

        return train_graphs, test_graphs, label_names


class ModelNet10Dataset(ModelNetDataset):
    def __init__(self, dataset_root_path=None, num_processes=50):
        super().__init__(
            dataset_name="ModelNet10",
            download_urls="http://vision.princeton.edu/projects/2014/3DShapeNets/ModelNet10.zip",
            dataset_root_path=dataset_root_path,
            num_processes=num_processes
        )


class ModelNet40Dataset(ModelNetDataset):
    def __init__(self, dataset_root_path=None, num_processes=50):
        super().__init__(
            dataset_name="ModelNet40",
            download_urls="http://modelnet.cs.princeton.edu/ModelNet40.zip",
            dataset_root_path=dataset_root_path,
            num_processes=num_processes
        )



# coding=utf-8
import tensorflow as tf
from tf_geometric.data.graph import BatchGraph
from tf_geometric.nn.pool.topk_pool import topk_pool


def sort_pool(x, edge_index, edge_weight, node_graph_index,
              k=None, ratio=None,
              sort_index=-1, training=None):
    """
    Functional API for SortPool "An End-to-End Deep Learning Architecture for Graph Classification"

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param node_graph_index: Tensor/NDArray, shape: [num_nodes], graph index for each node
    :param k: Keep top k targets for each source
    :param ratio: Keep num_targets * ratio targets for each source
    :param sort_index: The sort_index_th index of the last axis will used for sort.
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
    """

    score = x[:, sort_index]
    topk_node_index = topk_pool(node_graph_index, score, k=k, ratio=ratio)

    pooled_graph = BatchGraph(
        x=x,
        edge_index=edge_index,
        node_graph_index=node_graph_index,
        edge_graph_index=None,
        edge_weight=edge_weight
    ).sample_new_graph_by_node_index(topk_node_index)

    return pooled_graph.x, pooled_graph.edge_index, pooled_graph.edge_weight, pooled_graph.node_graph_index


# coding=utf-8

import numpy as np
from tf_geometric.data.dataset import DownloadableDataset
import os
import scipy.io as scio


class MultiLabelBlogCatalogDataset(DownloadableDataset):

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_name="MultiLabelBlogCatalog",
                         download_urls=[
                             "https://github.com/CrawlScript/gnn_datasets/raw/master/BlogCatalog/multi_label_blog_catalog.zip",
                             "http://cdn.zhuanzhi.ai/github/multi_label_blog_catalog.zip"
                         ],
                         download_file_name="multi_label_blog_catalog.zip",
                         cache_name="cache.p",
                         dataset_root_path=dataset_root_path,
                         )

    def process(self):
        data_path = os.path.join(self.raw_root_path, "multi_label_blog_catalog.mat")
        data = scio.loadmat(data_path)

        adj = data['network'].tocoo()
        edge_index = np.stack([adj.row, adj.col], axis=0)

        y = data['group'].tocoo().toarray().astype(np.float32)

        return edge_index, y



# coding=utf-8
import warnings
from tf_geometric.nn.conv.sgc import sgc
from tf_geometric.nn.conv.gcn import gcn_build_cache_for_graph
import tensorflow as tf


class SGC(tf.keras.Model):
    """
    The simple graph convolutional operator from the `"Simplifying Graph
    Convolutional Networks" <https://arxiv.org/abs/1902.07153>`_ paper
    """

    def __init__(self, units, k=1, activation=None, use_bias=True, renorm=True, improved=False,
                 kernel_regularizer=None, bias_regularizer=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        """
        :param units: Size of each output sample..
        :param k: Number of hops.(default: :obj:`1`)
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """

        self.units = units
        self.use_bias = use_bias
        self.renorm = renorm
        self.improved = improved
        self.k = k
        self.activation = activation
        self.kernel = None
        self.bias = None

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def build(self, input_shape):

        x_shape = input_shape[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        gcn_build_cache_for_graph(graph, self.renorm, self.improved, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn("'SGC.cache_normed_edge(graph, override)' is deprecated, use 'SGC.build_cache_for_graph(graph, override)' instead", DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)


    def call(self, inputs, cache=None, training=None, mask=None):
        """
        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, num_units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return sgc(x, edge_index, edge_weight, self.k, self.kernel,
                   bias=self.bias, activation=self.activation,
                   renorm=self.renorm, improved=self.improved, cache=cache)



# coding=utf-8

import numpy as np
import os
from tqdm import tqdm
import networkx as nx
from tf_geometric.data.graph import Graph
from tf_geometric.data.dataset import DownloadableDataset
import json

from tf_geometric.utils.data_utils import load_cache
from tf_geometric.utils.graph_utils import convert_edge_to_directed


class TUDataset(DownloadableDataset):

    def __init__(self, dataset_name, dataset_root_path=None):
        tu_base_url = 'https://ls11-www.cs.tu-dortmund.de/people/morris/graphkerneldatasets'
        super().__init__(dataset_name=dataset_name,
                         download_urls=[
                             "{}/{}.zip".format(tu_base_url, dataset_name)
                         ],
                         download_file_name="{}.zip".format(dataset_name),
                         cache_name="{}.p".format(dataset_name),
                         dataset_root_path=dataset_root_path,
                         )
        self.txt_root_path = os.path.join(self.raw_root_path, self.dataset_name)
        self.prefix = "{}_".format(self.dataset_name)

    def _build_label_id_index_dict(self, label_ids):
        sorted_label_ids = sorted(list(set(label_ids)))
        label_id_index_dict = {label_id: label_index for label_index, label_id in enumerate(sorted_label_ids)}
        return label_id_index_dict

    def _convert_label_ids_to_indices(self, label_ids):
        label_id_index_dict = self._build_label_id_index_dict(label_ids)
        return np.array([label_id_index_dict[label_id] for label_id in label_ids])

    def process(self):

        print("reading graph_indicator...")
        node_graph_index = self.read_txt_as_array("graph_indicator", dtype=np.int32)
        graph_index_offset = node_graph_index.min()
        node_graph_index -= graph_index_offset
        print("reading edges...")
        edges = self.read_txt_as_array("A", dtype=np.int32) - graph_index_offset
        edge_graph_index = node_graph_index[edges[:, 0]]
        num_graphs = node_graph_index.max() + 1

        # node_labels
        node_label_ids = self.read_txt_as_array("node_labels", dtype=np.int32)
        if node_label_ids is not None:
            node_labels = self._convert_label_ids_to_indices(node_label_ids)
        else:
            node_labels = None

        # node_labels
        edge_label_ids = self.read_txt_as_array("edge_labels", dtype=np.int32)
        if edge_label_ids is not None:
            edge_labels = self._convert_label_ids_to_indices(edge_label_ids)
        else:
            edge_labels = None

        # node_labels
        node_attributes_list = self.read_txt_as_array("node_attributes", dtype=np.float32)

        # graph_labels
        graph_label_ids = self.read_txt_as_array("graph_labels", dtype=np.int32)
        if graph_label_ids is not None:
            graph_labels = self._convert_label_ids_to_indices(graph_label_ids)
        else:
            graph_labels = None

        def create_empty_graph():
            graph = {"edge_index": []}

            if node_labels is not None:
                graph["node_labels"] = []

            if node_attributes_list is not None:
                graph["node_attributes"] = []

            if edge_labels is not None:
                graph["edge_labels"] = []

            if graph_labels is not None:
                graph["graph_label"] = None

            return graph

        graphs = [create_empty_graph() for _ in range(num_graphs)]

        start_node_index = np.full([num_graphs], -1).astype(np.int32)
        for node_index, graph_index in enumerate(node_graph_index):
            if start_node_index[graph_index] < 0:
                start_node_index[graph_index] = node_index

        for graph_index, graph in enumerate(graphs):
            end_index = start_node_index[graph_index + 1] if graph_index < num_graphs - 1 else len(node_graph_index)
            num_nodes = end_index - start_node_index[graph_index]
            graph["num_nodes"] = num_nodes

        # edge_index
        for graph_index, edge in zip(edge_graph_index, edges):
            graph = graphs[graph_index]
            graph["edge_index"].append(edge)

        if node_labels is not None:
            # node_labels -= node_labels.min()
            for graph_index, node_label in zip(node_graph_index, node_labels):
                graph = graphs[graph_index]
                graph["node_labels"].append(node_label)

        if edge_labels is not None:
            # edge_labels -= edge_labels.min()
            for graph_index, edge_label in zip(edge_graph_index, edge_labels):
                graph = graphs[graph_index]
                graph["edge_labels"].append(edge_label)


        if node_attributes_list is not None:
            node_attributes_list = node_attributes_list.reshape(node_attributes_list.shape[0], -1)
            for graph_index, node_attributes in zip(node_graph_index, node_attributes_list):
                graph = graphs[graph_index]
                graph["node_attributes"].append(node_attributes)


        # graph_labels

        if graph_labels is not None:
            # graph_labels -= graph_labels.min()
            for graph, graph_label in zip(graphs, graph_labels):
                graph["graph_label"] = np.array([graph_label]).astype(np.int32)

        for i, graph in enumerate(graphs):
            edge_index = np.array(graph["edge_index"]).T - start_node_index[i]
            graph["edge_index"] = edge_index

            if node_labels is not None:
                graph["node_labels"] = np.array(graph["node_labels"]).astype(np.int32)

            if node_attributes_list is not None:
                graph["node_attributes"] = np.array(graph["node_attributes"]).astype(np.float32)

            if edge_labels is not None:
                graph["edge_labels"] = np.array(graph["edge_labels"]).astype(np.int32)

            num_nodes = graph["num_nodes"]
            nx_graph = nx.Graph()
            nx_graph.add_nodes_from(np.arange(num_nodes))
            nx_graph.add_edges_from(edge_index.T)
            degrees = np.array([nx_graph.degree(node_index) for node_index in range(num_nodes)]).astype(np.int32)
            graph["degrees"] = degrees

        return graphs


    def get_path_by_fid(self, fid):
        fname = "{}_{}.txt".format(self.dataset_name, fid)
        return os.path.join(self.txt_root_path, fname)


    def read_txt_as_array(self, fid, dtype):
        path = self.get_path_by_fid(fid)
        if not os.path.exists(path):
            return None
        data_list = []
        with open(path, "r", encoding="utf-8") as f:
            for line in tqdm(f):
                line = line.strip()
                if len(line) == 0:
                    continue
                items = line.split(",")
                items = [dtype(item) for item in items]
                data = items[0] if len(items) == 1 else items
                data_list.append(data)
        data_list = np.array(data_list).astype(dtype)
        return data_list







# coding=utf-8

import tensorflow as tf
from tf_sparse import SparseMatrix

from tf_geometric.sparse.sparse_adj import SparseAdj
from tf_geometric.utils.graph_utils import remove_self_loop_edge, get_laplacian, LaplacianMaxEigenvalue
import tf_sparse as tfs

CACHE_KEY_CHEBYNET_NORMED_EDGE_TEMPLATE = "chebynet_normed_edge_{}"


def compute_cache_key(normalization_type):
    """
    Compute the cached key based on GCN normalization configurations: renorm and improved

    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :return: The corresponding cached key for the given GCN normalization configuration.
    """
    return CACHE_KEY_CHEBYNET_NORMED_EDGE_TEMPLATE.format(normalization_type)


# def chebynet_compute_lambda_max(edge_index, edge_weight, normalization_type, num_nodes, cache=None):
#     if cache is not None:
#         cache_key = "chebynet_lambda_max_{}".format(normalization_type)
#         cached_data = cache.get(cache_key, None)
#         if cached_data is not None:
#             return cached_data
#
#     lambda_max = LaplacianMaxEigenvalue(edge_index, edge_weight, num_nodes)(normalization_type=normalization_type)
#
#     if cache is not None:
#         cache[cache_key] = lambda_max
#
#     return lambda_max


def chebynet_norm_edge(edge_index, num_nodes, edge_weight=None, normalization_type="sym", use_dynamic_lambda_max=False, cache=None):
    if cache is not None:
        cache_key = compute_cache_key(normalization_type)
        cached_data = cache.get(cache_key, None)
        if cached_data is not None:
            return cached_data

    edge_index, edge_weight = remove_self_loop_edge(edge_index, edge_weight)

    updated_edge_index, updated_edge_weight = get_laplacian(edge_index, num_nodes, edge_weight, normalization_type)

    # lambda_max = chebynet_compute_lambda_max(edge_index, edge_weight, normalization_type, num_nodes, cache=cache)
    if use_dynamic_lambda_max:
        lambda_max = LaplacianMaxEigenvalue(edge_index, num_nodes, edge_weight)(normalization_type=normalization_type)
    else:
        lambda_max = 2.0
    scaled_edge_weight = (2.0 * updated_edge_weight) / lambda_max

    assert edge_weight is not None

    if cache is not None:
        cache[cache_key] = updated_edge_index, scaled_edge_weight

    return updated_edge_index, scaled_edge_weight


def chebynet_cache_normed_edge(graph, normalization_type="sym", use_dynamic_lambda_max=False, override=False):
    """
    Manually compute the normed edge based on the given GCN normalization configuration (renorm and improved) and put it in graph.cache.
    If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

    :param graph: tfg.Graph, the input graph.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param override: Whether to override existing cached normed edge.
    :return: None
    """
    if override:
        cache_key = compute_cache_key(normalization_type)
        graph.cache[cache_key] = None
    chebynet_norm_edge(graph.edge_index, graph.num_nodes, graph.edge_weight, normalization_type,
                       use_dynamic_lambda_max=use_dynamic_lambda_max, cache=graph.cache)


def chebynet(x, edge_index, edge_weight, k, kernels, bias=None, activation=None, normalization_type="sym", use_dynamic_lambda_max=False, cache=None):
    num_nodes = tfs.shape(x)[0]
    # lambda_max = chebynet_compute_lambda_max(x, edge_index, edge_weight, normalization_type, cache=cache)

    num_edges = tf.shape(edge_index)[1]
    if edge_weight is None:
        edge_weight = tf.ones([num_edges], dtype=tf.float32)

    normed_edge_index, normed_edge_weight = chebynet_norm_edge(edge_index, num_nodes, edge_weight, normalization_type,
                                                           use_dynamic_lambda_max=use_dynamic_lambda_max, cache=cache)
    normed_sparse_adj = SparseAdj(normed_edge_index, normed_edge_weight, [num_nodes, num_nodes])

    if isinstance(x, SparseMatrix):
        x = x.to_dense()

    T0_x = x

    if isinstance(T0_x, tf.sparse.SparseTensor):
        out = tf.sparse.sparse_dense_matmul(T0_x, kernels[0])
    else:
        out = T0_x @ kernels[0]

    if k > 1:
        # T1_x = aggregate_neighbors(x, norm_edge_index, norm_edge_weight, gcn_mapper, sum_reducer, identity_updater)
        if isinstance(x, tf.sparse.SparseTensor):
            dense_x = tf.sparse.to_dense(x)
        else:
            dense_x = x

        T1_x = normed_sparse_adj @ dense_x
        h = T1_x @ kernels[1]
        out += h

    if k > 2:

        if isinstance(T0_x, tf.sparse.SparseTensor):
            T0_x = tf.sparse.to_dense(T0_x)

        for i in range(2, k):
            # T2_x = aggregate_neighbors(T1_x, norm_edge_index, norm_edge_weight, gcn_mapper, sum_reducer,
            #                            identity_updater)  ##L^T_{k-1}(L^)

            T2_x = normed_sparse_adj @ T1_x * 2.0 - T0_x
            h = T2_x @ kernels[i]
            out += h

            T0_x, T1_x = T1_x, T2_x

    if bias is not None:
        out += bias

    if activation is not None:
        out = activation(out)

    return out



# coding=utf-8


"""
Sparse Adj for Computation
"""
from tf_sparse import SparseMatrix

from tf_geometric.nn.kernel.segment import segment_softmax
from tf_geometric.utils.graph_utils import add_self_loop_edge, remove_self_loop_edge


class SparseAdj(SparseMatrix):

    def __init__(self, index, value=None, shape=None, merge=False,
                 edge_weight=None):
        if value is not None and edge_weight is not None:
            raise Exception("\"edge_weight\" is an alias for \"value\", and it is deprecated. You should only provide \"value\".")
        super().__init__(index, value, shape, merge)

    @property
    def edge_index(self):
        return self.index

    @property
    def edge_weight(self):
        return self.value

    def add_self_loop(self, fill_weight=1.0):
        num_nodes = self._shape[0]
        updated_edge_index, updated_edge_weight = add_self_loop_edge(self.index, num_nodes,
                                                                     edge_weight=self.value,
                                                                     fill_weight=fill_weight)
        return self.__class__(updated_edge_index, updated_edge_weight, self._shape)

    def remove_self_loop(self):
        updated_edge_index, updated_edge_weight = remove_self_loop_edge(self.index, edge_weight=self.value)
        return self.__class__(updated_edge_index, updated_edge_weight, self._shape)



# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.conv.le_conv import le_conv


class LEConv(tf.keras.Model):
    """
    Graph Convolutional Layer
    """

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.self_kernel = self.add_weight("self_kernel", shape=[num_features, self.units],
                                           initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.self_use_bias:
            self.self_bias = self.add_weight("self_bias", shape=[self.units],
                                             initializer="zeros", regularizer=self.bias_regularizer)

        self.aggr_self_kernel = self.add_weight("aggr_self_kernel", shape=[num_features, self.units],
                                                initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.aggr_self_use_bias:
            self.aggr_self_bias = self.add_weight("aggr_self_bias", shape=[self.units],
                                                  initializer="zeros", regularizer=self.bias_regularizer)

        self.aggr_neighbor_kernel = self.add_weight("aggr_neighbor_kernel", shape=[num_features, self.units],
                                                    initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.aggr_neighbor_use_bias:
            self.aggr_neighbor_bias = self.add_weight("aggr_neighbor_bias", shape=[self.units],
                                                      initializer="zeros", regularizer=self.bias_regularizer)

    def __init__(self, units, activation=None,
                 self_use_bias=True,
                 aggr_self_use_bias=True,
                 aggr_neighbor_use_bias=False,
                 kernel_regularizer=None, bias_regularizer=None, *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units

        self.activation = activation

        self.self_use_bias = self_use_bias
        self.aggr_self_use_bias = aggr_self_use_bias
        self.aggr_neighbor_use_bias = aggr_neighbor_use_bias

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.self_kernel = None
        self.self_bias = None
        self.aggr_self_kernel = None
        self.aggr_self_bias = None
        self.aggr_neighbor_kernel = None
        self.aggr_neighbor_bias = None

    def call(self, inputs, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return le_conv(x, edge_index, edge_weight,
                       self.self_kernel, self.self_bias,
                       self.aggr_self_kernel, self.aggr_self_bias,
                       self.aggr_neighbor_kernel, self.aggr_neighbor_bias,
                       activation=self.activation)



# coding=utf-8
import os

from tf_sparse import SparseMatrix

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
from tf_geometric.utils import tf_utils
import tensorflow as tf
import tf_geometric as tfg
from tqdm import tqdm
import time
import tf_sparse as tsp

graph, (train_index, valid_index, test_index) = tfg.datasets.CoraDataset().load_data()
graph.x = tsp.eye(graph.num_nodes)
# graph.x = tf.sparse.eye(graph.num_nodes)

num_classes = graph.y.max() + 1
drop_rate = 0.5
learning_rate = 1e-2


# Multi-layer GCN Model
class GCNModel(tf.keras.Model):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gcn0 = tfg.layers.GCN(16, activation=tf.nn.relu)
        self.gcn1 = tfg.layers.GCN(num_classes)
        self.dropout = tsp.layers.Dropout(drop_rate)

    def call(self, inputs, training=None, mask=None, cache=None):
        x, edge_index, edge_weight = inputs
        h = self.dropout(x, training=training)

        # if isinstance(x, SparseMatrix):
        #     h = x.dropout(drop_rate, training=training)
        # else:
        #     h = self.dropout(x, training=training)

        h = self.gcn0([h, edge_index, edge_weight], cache=cache)
        h = self.dropout(h, training=training)
        h = self.gcn1([h, edge_index, edge_weight], cache=cache)
        return h


model = GCNModel()


# @tf_utils.function can speed up functions for TensorFlow 2.x.
# @tf_utils.function is not compatible with TensorFlow 1.x and dynamic graph.cache.
@tf_utils.function
def forward(graph, training=False):
    return model([graph.x, graph.edge_index, graph.edge_weight], training=training, cache=graph.cache)


# The following line is only necessary for using GCN with @tf_utils.function
# For usage without @tf_utils.function, you can commont the following line and GCN layers can automatically manager the cache
model.gcn0.build_cache_for_graph(graph)


@tf_utils.function
def compute_loss(logits, mask_index, vars):
    masked_logits = tf.gather(logits, mask_index)
    masked_labels = tf.gather(graph.y, mask_index)
    losses = tf.nn.softmax_cross_entropy_with_logits(
        logits=masked_logits,
        labels=tf.one_hot(masked_labels, depth=num_classes)
    )

    kernel_vars = [var for var in vars if "kernel" in var.name]
    l2_losses = [tf.nn.l2_loss(kernel_var) for kernel_var in kernel_vars]

    return tf.reduce_mean(losses) + tf.add_n(l2_losses) * 5e-4


optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)


@tf_utils.function
def train_step():
    with tf.GradientTape() as tape:
        logits = forward(graph, training=True)
        loss = compute_loss(logits, train_index, tape.watched_variables())

    vars = tape.watched_variables()
    grads = tape.gradient(loss, vars)
    optimizer.apply_gradients(zip(grads, vars))
    return loss


@tf_utils.function
def evaluate():
    logits = forward(graph)
    masked_logits = tf.gather(logits, test_index)
    masked_labels = tf.gather(graph.y, test_index)

    y_pred = tf.argmax(masked_logits, axis=-1, output_type=tf.int32)

    corrects = tf.equal(y_pred, masked_labels)
    accuracy = tf.reduce_mean(tf.cast(corrects, tf.float32))
    return accuracy


for step in range(1, 201):
    loss = train_step()
    if step % 20 == 0:
        accuracy = evaluate()
        print("step = {}\tloss = {}\taccuracy = {}".format(step, loss, accuracy))

print("\nstart speed test...")
num_test_iterations = 1000
start_time = time.time()
for _ in tqdm(range(num_test_iterations)):
    logits = forward(graph)
end_time = time.time()
print("mean forward time: {} seconds".format((end_time - start_time) / num_test_iterations))

if tf.__version__[0] == "1":
    print("** @tf_utils.function is disabled in TensorFlow 1.x. "
          "Upgrade to TensorFlow 2.x for 10X faster speed. **")
